@@ -79,18 +79,19 @@ class Country(object):
 
         if month_id is None:
             try:
-                return int(descriptors.sort_values(by='month_end',
-                                                   ascending=False,
-                                                   ignore_index=True).loc[0].id)
+                x = int(descriptors.sort_values(by='month_end',
+                                                ascending=False,
+                                                ignore_index=True).loc[0].id)
+                return x
             except ValueError:
-                raise ValueError('Country does not exist!')
+                raise ValueError("Country does not exist")
         else:
             descriptors = descriptors[(descriptors.month_start <= month_id) &
-                                      (descriptors.month_end > month_id)]
+                                      (descriptors.month_end >= month_id)]
         try:
             return int(max(descriptors.id))
         except ValueError:
-            raise ValueError('Country does not exist at this point in time!')
+            raise ValueError(f"Country with {name_var} = {name_value} does not exist at month {month_id}!")
 
 
     @staticmethod
